//---------------------------------------------------------------------------

#pragma hdrstop

#include <fstream.h>
#include "Landscape.h"

int Landscape::LandCounter = 0;

#if SEED
int32_t Lseed = (int32_t) SEED;
#else
int32_t Lseed = (int32_t)time(0);
#endif
StochasticLib1 sto1land(Lseed);

ofstream outspatial("SpatialSumm.csv"); // spatial summary data file
bool write_outspatial_headers = 1;

//---------------------------------------------------------------------------

#pragma package(smart_init)

//------------------------------------------------------------------------------
//FUNCTION:   Square::Square
//ARGUMENTS:  none
//RETURNS:    none
//NOTES:      Square class default Constructor
//------------------------------------------------------------------------------

Square::Square(int x, int y)
{
sq.x = x; sq.y = y;
Ki = 0; Kit = 0; Dispersers = Residents = totDispersers = totResidents = Ngens = 0;
}

//------------------------------------------------------------------------------
//FUNCTION:   Square::various
//ARGUMENTS:  none
//RETURNS:    various
//NOTES:      Square functions to return individual attributes
//------------------------------------------------------------------------------

locn Square::get_locn(void) { return sq; }

int Square::get_Ki(void) { return Ki; }
int Square::get_Kit(void) { return Kit; }
int Square::get_PredDens(void) { return Dispersers+Residents; }
int Square::get_Dispersers(void) { return Dispersers; }
int Square::get_Residents(void) { return Residents; }

double Square::get_PredDisp(void) {
if (Dispersers == 0 && Residents == 0) return 0.0;
else return (double)Dispersers/(double)(Dispersers+Residents);
}

void Square::set_Ki(int k, int method, double sd)
// set the mean and current carrying capacities
{
Ki = k;
if (method == 0) Kit = sto1land.Poisson(Ki);
if (method == 1) Kit = sto1land.Normal(Ki,sd);
if (Kit < parameters->getKmin()) Kit = parameters->getKmin();
}

void Square::update_Ki(int method, double sd)
// update the mean and current carrying capacities
{
if (method == 0) Kit = sto1land.Poisson(Ki);
if (method == 1) Kit = sto1land.Normal(Ki,sd);
if (Kit < parameters->getKmin()) Kit = parameters->getKmin();
}

void Square::incrPredDens(int g) { if (g > 0) Dispersers++; else Residents++; }
void Square::incrTotals(void)
{ totDispersers += Dispersers; totResidents += Residents; Ngens++; }
void Square::resetPredStats(void) { Dispersers = Residents = 0; }

sqStats Square::get_sqStats(void)
{
sqStats s;
s.K = Ki;
if (Ngens < 1) {
  s.meanDisp = s.meanRes = 0.0;
}
else {
  s.meanDisp = (double)totDispersers / (double)Ngens;
  s.meanRes = (double)totResidents / (double)Ngens;
}
return s;
}

//------------------------------------------------------------------------------
//FUNCTION: Landscape::Landscape
//ARGUMENTS: None
//RETURNS: None
//NOTES: Landscape class constructor
//------------------------------------------------------------------------------

Landscape::Landscape()
{
int x,y;
xmax = parameters->getXmax(); ymax = parameters->getYmax();
int k;
double mu,sspatial,kdouble;

mu = parameters->getMU(); sspatial = parameters->getSdSpatial();
#if DEBUG
cout<<endl<<"Landscape: xmax "<<xmax<<" ymax "<<ymax
  <<" mu "<<mu<<" sspatial "<<sspatial<<" Kmin "<<parameters->getKmin()
  <<endl<<endl;
#endif

id = LandCounter; LandCounter++; //Unique identifier for each landscape
Kmean = Ktot = 0;
Kmin = 999999999;

if (write_outspatial_headers) {
  outspatial<<"Repl,Y,X,K,MeanNdisp,MeanNres"<<endl;
  write_outspatial_headers = 0;
}
try {
for (y = ymax-1; y > -1; y--) {
  for (x = 0; x < xmax; x++) {
	Squares.push_back(new Square(x,y));
	kdouble = sto1land.Normal(mu,sspatial);
	k = (int)(kdouble+0.5);
  if (k < parameters->getKmin()) k = parameters->getKmin(); // ***** NEW LINE TO PREVENT SYSTEM CRASH *****
	locate_square(x,y)->set_Ki(k, parameters->getTempMethod(), parameters->getSdTemp());
	Kmean += k; Ktot += k;
	if (k < Kmin) Kmin = k;
#if DEBUG
  cout<<"("<<x<<","<<y<<") Ki "<<locate_square(x,y)->get_Ki()
    <<" Kit "<<locate_square(x,y)->get_Kit()<<endl;
#endif
  }
}
}
catch(...)
{
	int fix = 8;
}

}

Landscape::~Landscape()
{
for (int i = 0; i < Squares.size(); i++) delete Squares[i];
Squares.clear();
}

//------------------------------------------------------------------------------
//FUNCTION: Landscape::locate_square
//ARGUMENTS: x,y
//RETURNS: Pointer to required square
//NOTES: Locates a Square object within the landscape given its co-ordinates
//------------------------------------------------------------------------------

Square* Landscape::locate_square(int x,int y)
{
return Squares[(ymax-1-y)*xmax + x];
}

//------------------------------------------------------------------------------
//FUNCTION: Landscape::get_bounds
//ARGUMENTS: none
//RETURNS: location
//NOTES: returns maximum X and Y of the landscape
//------------------------------------------------------------------------------

locn Landscape::get_bounds(void) { locn l; l.x = xmax; l.y = ymax; return l; }

int Landscape::get_id(void) { return id; }
int Landscape::get_Kmean(void) { return Kmean; }
int Landscape::get_Ktot(void) { return Ktot; }
int Landscape::get_Kmin(void) { return Kmin; }

//------------------------------------------------------------------------------
//FUNCTION: Landscape::update_K
//ARGUMENTS: method to be applied and temporal standard deviation
//RETURNS: none
//NOTES: updates carrying capacity in all squares of the landscape
//------------------------------------------------------------------------------

void Landscape::update_K(int method, double sd)
{

int i;
Ktot = 0;

i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      Squares[i]->update_Ki(method,sd);
      Ktot += Squares[i]->get_Kit();
    }
    i++;
  }
}

}

void Landscape::incrTotals(void)
{
int i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      Squares[i]->incrTotals();
    }
    i++;
  }
}

}

//------------------------------------------------------------------------------
void Landscape::resetPredStats()
{
int i;

i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      Squares[i]->resetPredStats();
    }
    i++;
  }
}

}

//------------------------------------------------------------------------------
int Landscape::get_min_K(void)
// Calculate minimum carrying capacity at current time-step
{

int i,k;
int kmin = 999999999;

i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      k = Squares[i]->get_Kit();
      if (k < kmin) kmin = k;
    }
    i++;
  }
}

return kmin;

}

int Landscape::TotalPredN(void)
// Calculate total number of beetles
{

int i;
int total = 0;

i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      total += (Squares[i]->get_Dispersers()+Squares[i]->get_Residents());
    }
    i++;
  }
}
return total;

}

void Landscape::spatialSummary(int repl)
// Write spatial summary statistics
{

int i;
sqStats s;

i = 0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      s = Squares[i]->get_sqStats();
      outspatial<<repl<<","<<y<<","<<x<<","<<s.K<<","<<s.meanDisp<<","<<s.meanRes
        <<endl;
    }
    i++;
  }
}

}


locn Landscape::randomSquare(double rr)
// finds a square at random in proportion to squares' mean carrying capacity
// i.e. a random lottery biased by carrying capacity
{
locn l;
double prop,cumprop;

int i = 0;
cumprop = 0.0;
for (int y = ymax-1; y > -1; y--) {
  for (int x = 0; x < xmax; x++) {
    if (Squares[i] != 0) {
      prop = (double)(Squares[i]->get_Ki()) / (double)Kmean;
      cumprop += prop;
      if (rr < cumprop) { // current square is chosen
        l.x = x; l.y = y;
        x = xmax+1; y = -2; // to break out of loops
      }
    }
    i++;
  }
}

return l;
}

locn Landscape::attractiveSquare(double rr,int option)
// finds a square at random in relation to the ratio of the current density to the
// current carrying, i.e. a random lottery biased by (proportional) available space
{
locn l;
double cumprop,p,e;
std::vector <double> prop;
int i,N,Nadj,K;

#if DEBUG
cout<<endl;
#endif
cumprop = 0.0;
for (i = 0; i < Squares.size() ; i++) {
  N = Squares[i]->get_PredDens();
  K = Squares[i]->get_Kit();
  if (K < 1) prop[i] = 0.0;
  else {
    switch (option) {
      case 0: // P in proportion to 1 - N/K
        if (N >= K) Nadj = K-1; // allow small probability of settlement if over-full
        else Nadj = N;
        p = 1.0 - ((double)Nadj/(double)K);
        break;
      case 1: // P in proportion to 1/exp(N/K)
        p = 1.0 / exp((double)N/(double)K);
        break;
      case 2: // P in proportion to exp((K-N)/K)/(1+exp((K-N)/K))
        e = exp((double)(K-N)/(double)K);
        p = e / (1.0 + e);
        break;
      default:
        p = 1.0;
    }
    prop.push_back(p);
  }
  cumprop += prop[i];
#if DEBUG
  l = Squares[i]->get_locn();
  cout<<"i "<<i<<" ("<<l.x<<","<<l.y<<") N "<<N<<" K "<<K
    <<" prop "<<prop[i]<<" cumprop "<<cumprop<<endl;
#endif
}
for (i = 0; i < Squares.size() ; i++) {
  prop[i] /= cumprop;
  if (i > 0) prop[i] += prop[i-1];
#if DEBUG
  cout<<"i "<<i<<" prop "<<prop[i]<<endl;
#endif
}
#if DEBUG
cout<<"rr "<<rr<<endl;
#endif
for (i = 0; i < Squares.size() ; i++) {
  if (rr < prop[i]) {
    l = Squares[i]->get_locn();
    i = Squares.size()+1; // to break out of loop
  }
}

return l;
}


//------------------------------------------------------------------------------


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
