//---------------------------------------------------------------------------
                          
#pragma hdrstop

#include "Predators.h"

#if SEED
int32_t Pseed = (int32_t) SEED;
#else
int32_t Pseed = (int32_t)time(0);
#endif
StochasticLib1 sto1pred(Pseed);

#if BATCH
ofstream predbrd("predbrd.csv"); // predator breeding stats file
bool write_predbrd_headers = 1;
#endif

//---------------------------------------------------------------------------
#pragma package(smart_init)

//------------------------------------------------------//Constructor definition
PredPop::PredPop(int nn, double ll, double bb, double mm)
{
Nstart = nn; lambda = ll; b = bb; dispMort = mm;
#if BATCH
#if WRITEPREDBRD
if (write_predbrd_headers) {
  predbrd<<"Gen,Adult,ID,X,Y,preyDens,predDens,a,meanJuvs,juvs"<<endl;
  write_predbrd_headers = 0;
}
#endif
#endif
}

PredPop::~PredPop()
{
for (int i = 0; i < Adults.size(); i++) delete Adults[i];
Adults.clear();
for (int i = 0; i < Juvs.size(); i++) delete Juvs[i];
Juvs.clear();
}

double PredPop::get_lambda() { return lambda; }
double PredPop::get_b() { return b; }
double PredPop::get_dispMort() { return dispMort; }

//------------------------------------------------//Sets initial population
void PredPop::InitialisePreds(Landscape* pLand)
{

locn locn;

#if DEBUG
cout<<"***** Initial animals *****"<<endl;
#endif
for (int i = 0; i < Nstart; i++)
{
  // set pointer to new animal heap object into vector container
  Adults.push_back(new Animal);
  // increment predator count in square where animal created
  locn = Adults[i]->get_locn();
  pLand->locate_square(locn.x,locn.y)->incrPredDens(Adults[i]->get_genome().gene0);
#if DEBUG
  cout<<i<<" "<<Adults[i]<<" id "<<Adults[i]->get_id()
    <<" wing "<<Adults[i]->get_genome().gene0<<" at ("<<locn.x<<","<<locn.y<<")"<<endl;
#endif
}

}


//------------------------------------------------------------------------------
int PredPop::Breed(Landscape* pLand, double Pmut, int gen)
{        
double predDensity,q,meanJuvs;
int K,juvs;
int Njuvs = 0;
locn locn;
Square* sq;

#if DEBUG
cout<<endl<<"Breed: gen "<<gen<<" N animals = "<<Adults.size()<<endl;
#endif

for (int i = 0; i < Adults.size(); i++) { // for each adult in the population
  locn = Adults[i]->get_locn();
  sq = pLand->locate_square(locn.x,locn.y);
  K = sq->get_Kit(); // get the current carrying capacity of the square
  if (K < 1) K = 1;
  predDensity = sq->get_PredDens(); // get the current density
  // expected no. of juveniles from Beverton-Holt model
  q = (pow(lambda,(1.0/b)) - 1.0) / K;
  meanJuvs = lambda * pow((1 + q*predDensity),-b);
  // ranodm no. of juveniles from Poisson distribn. having B-H mean
  if (meanJuvs <= 0.0) juvs = 0;
  else juvs = sto1pred.Poisson(meanJuvs);
  Njuvs += juvs;
#if DEBUG
  cout<<"Adult: "<<i<<" id "<<Adults[i]->get_id()
    <<" at ("<<locn.x<<","<<locn.y<<") K "<<K
    <<" dens "<<predDensity<<" q "<<q<<" meanJuvs "<<meanJuvs<<" juvs "<<juvs<<endl;
#endif
#if BATCH
#if WRITEPREDBRD
  predbrd<<gen<<","<<i<<","<<Adults[i]->get_id()<<","<<locn.x<<","<<locn.y<<","
    <<K<<","<<predDensity<<","
    <<q<<","<<meanJuvs<<","<<juvs<<endl;
#endif
#endif
  for (int j = 0; j < juvs; j++) {
    // create new animal objects on the heap using copy constructor and pointer to
    // parent object, pointers then stored in juvenile vector
    Juvs.push_back(new Animal(Adults[i],Pmut));
  }
  delete Adults[i]; // delete adult heap objects once juvs have been created
}
Adults.clear(); // remove all pointers to adult heap objects avoiding dangling pointers
return Njuvs; // return total number of juveniles produced
}


//------------------------------------------------------------------------------
void PredPop::Disperse(Landscape* pLand)
{
locn locn;
genome g;
Square* sq;

#if DEBUG
cout<<endl<<"Disperse:"<<" N juvs = "<<Juvs.size()<<endl<<endl;
#endif

// randomise order in which juveniles are processed prior to dispersal
random_shuffle (Juvs.begin(), Juvs.end());

// all residents need to be processed prior to any dispersers to allow for
// dispersal method 2 (fn of N/K)
for (int i = 0; i < Juvs.size(); i++) {
  g = Juvs[i]->get_genome();
  if (g.gene0 == 0) { // short-winged resident
    locn = Juvs[i]->get_locn();
#if DEBUG
    cout<<"i "<<i<<" id "<<Juvs[i]->get_id()
      <<" wing "<<g.gene0<<" at ("<<locn.x<<","<<locn.y<<")";
#endif
    sq = pLand->locate_square(locn.x,locn.y);
    sq->incrPredDens(g.gene0);
    Adults.push_back(Juvs[i]);
#if DEBUG
    cout<<" remained at ("<<locn.x<<","<<locn.y<<") Kit = "<<sq->get_Kit()<<endl;
#endif
  }
}

// now process dispersers
for (int i = 0; i < Juvs.size(); i++) {
  g = Juvs[i]->get_genome();
  if (g.gene0 == 1) { // long-winged disperser
#if DEBUG
    locn = Juvs[i]->get_locn();
    cout<<"i "<<i<<" id "<<Juvs[i]->get_id()
      <<" wing "<<g.gene0<<" at ("<<locn.x<<","<<locn.y<<")";
#endif
    locn = Juvs[i]->Move(pLand,dispMort);
    if (locn.x < -900) { // juvenile has died in dispersal
#if DEBUG
      cout<<" ** died during dispersal"<<endl;
#endif
      delete Juvs[i];
    }
    else {
      sq = pLand->locate_square(locn.x,locn.y);
      sq->incrPredDens(g.gene0);
      Adults.push_back(Juvs[i]);
#if DEBUG
      cout<<" ** moved to ("<<locn.x<<","<<locn.y<<")"
        <<" Ki = "<<sq->get_Ki()<<" Kit = "<<sq->get_Kit()<<endl;
#endif
    }
  }
}

Juvs.clear();
}

//------------------------------------------------------------------------------
PredStats PredPop::GetStats(Landscape* pLand)
{
// Calculates statistics across the whole population:
// total number of cells occupied
// mean, s.d., minimum and maximum dispersal probability
// mean stopping probability parameters (NOT USED HERE)
// mean no. of steps taken by dispersers

PredStats s;
genome g;
double Pdisp = 0.0;
double PstopInt = 0.0;
double PstopSlope = 0.0;
double Pdispsq = 0.0;
int sumNsteps = 0;
int Nsteps;
int dispersers = 0;
int residents = 0;

s.n = Adults.size();
s.minDisp = 1.0;
s.maxDisp = 0.0;
s.maxNsteps = 0;

// census population
for (int i = 0; i < Adults.size(); i++) {
//  Pdisp = Adults[i]->get_Pdisperse();
  g = Adults[i]->get_genome();
  Pdisp += g.gene0; Pdispsq += g.gene0*g.gene0;
  if (g.gene0 < s.minDisp) s.minDisp = g.gene0;
  if (g.gene0 > s.maxDisp) s.maxDisp = g.gene0;
//  PstopInt += g.gene1; PstopSlope += g.gene2;
  Nsteps = Adults[i]->get_Nsteps();
  if (g.gene0 > 0) {
    sumNsteps += Nsteps; dispersers++;
    if (Nsteps > s.maxNsteps) s.maxNsteps = Nsteps;
  }
  else residents++;
}

// calculate stats
if (Adults.size() > 0) {
  s.meanDisp = Pdisp / (double)Adults.size();
//  s.meanPstopInt = PstopInt / (double)Adults.size();
//  s.meanPstopSlope = PstopSlope / (double)Adults.size();
}
else s.meanDisp = 0.0;
if (Adults.size() < 2) s.sdDisp = 0.0;
else {
  // the following contortion is required to prevent a domain error in sqrt()
  s.sdDisp = Pdispsq/(double)Adults.size() - s.meanDisp*s.meanDisp;
  if (s.sdDisp <= 0.0) s.sdDisp = 0.0;
  else s.sdDisp = sqrt(s.sdDisp);
}
if (dispersers > 0) {
  s.meanNsteps = (double)sumNsteps / (double)dispersers;
}
else s.meanNsteps = 0.0;

s.dispersers = dispersers;
s.residents = residents;

return s;
}

//------------------------------------------------------------------------------
 

